#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <math.h>
#include <algorithm>

#include "TypeDef.h"
#include "TComLut3D.h"

using namespace std;

// static functions : ======================================================
/*
static  Void  my_assert(char *comment, Bool bTest)
{
  if ( !bTest ) {
    printf("assertion failed:\t%s\n", comment);
    exit(0);
  }
}
*/
static  Int getNbLayer( Int nbp, Int nbLayerMax )
{
  Int iLayer=0;
  while ( iLayer < nbLayerMax ) {
    if ( (1 << iLayer) == (nbp-1) )
      return(iLayer+1);
    else
      iLayer++;
  }
  //assert("getNbLayer():\tNBP should be a power of 2 < 256", iLayer<NB_LAYER_MAX);

  return(-1);
}

static Void  copyNode( node *nodeDst, node *nodeSrc )
{
  nodeDst->sY = nodeSrc->sY;
  nodeDst->sU = nodeSrc->sU;
  nodeDst->sV = nodeSrc->sV;
}

static Int getNodeError( node *nodeRef, node *nodePred )
{
  node  nodeDiff;

  nodeDiff.sY = nodeRef->sY - nodePred->sY;
  nodeDiff.sU = nodeRef->sU - nodePred->sU;
  nodeDiff.sV = nodeRef->sV - nodePred->sV;

  // Quadratic error :
  //return( nodeDiff.sY*nodeDiff.sY + nodeDiff.sU*nodeDiff.sU + nodeDiff.sV*nodeDiff.sV );


  // Max Absolute error :
  nodeDiff.sY = abs( nodeDiff.sY );
  nodeDiff.sU = abs( nodeDiff.sU );
  nodeDiff.sV = abs( nodeDiff.sV );

  Short sTmp  = max( nodeDiff.sU, nodeDiff.sV );
  return( max( nodeDiff.sY, sTmp ) );
}

static  octTree *allocateOctTree(int level)
{
  octTree *pOctTree=NULL;

  if ( level ) {
    pOctTree = (octTree *)malloc(sizeof(octTree));
    pOctTree->bSplit = false;

    for ( int i=0 ; i<8 ; i++ )
      pOctTree->leave[i] = allocateOctTree( level - 1 );
  }

  return( pOctTree );
}

static  Void  deleteOctTree( octTree *pOctTree )
{
  if ( pOctTree ) {
    for ( int i=0 ; i<8 ; i++ ) {
      deleteOctTree( (octTree *)pOctTree->leave[i] );
      pOctTree->leave[i] = NULL;
    }

    free( pOctTree );
  }
}

//==========================================================================

TComLut3D::TComLut3D( Int nbp, Bool bAllocateOctree )
{
  m_nbp = nbp;
#if FIXED_ARRAY
  assert("nbp != >NBP", nbp==NBP);
#else
  m_yuvLut  = (node *)malloc(sizeof(node)*m_nbp*m_nbp*m_nbp);
#endif

#if USE_SHIFT_DIV
  m_dr  = BOX_MAX / (m_nbp-1);
  m_dg  = BOX_MAX / (m_nbp-1);
  m_db  = BOX_MAX / (m_nbp-1);

  Int dx = BOX_MAX / (m_nbp-1);
  m_shift_div = (Int)(log((double)dx)/log(2.0));
  if ( dx != (Int)pow(2.0,m_shift_div) ) {
    printf("TComLut3D:\tnbp should be power of 2.\n");
    exit(0);
  }
  m_shift_div3  = 3 * m_shift_div;
#endif

  if ( bAllocateOctree )  m_octTree = allocateOctTree( getNbLayer(m_nbp, 8) );
  else                    m_octTree = NULL;
}

TComLut3D::~TComLut3D()
{
#if FIXED_ARRAY
#else
  free(m_yuvLut);
  m_yuvLut = NULL;
#endif

  if ( m_octTree )  deleteOctTree( m_octTree );
}

#if USE_IBD
Void  TComLut3D::set_ibd( Int internalBitDepthY, Int internalBitDepthC )
{
#if USE_SHIFT_DIV
  m_dr  = BOX_MAX / (m_nbp-1);
  m_dg  = BOX_MAX / (m_nbp-1);
  m_db  = BOX_MAX / (m_nbp-1);
  m_dr  = m_dr  << (internalBitDepthY - 8);
  m_dg  = m_dg  << (internalBitDepthC - 8);
  m_db  = m_db  << (internalBitDepthC - 8);

  Int dx = BOX_MAX / (m_nbp-1);
  m_shift_div = (Int)(log((double)dx)/log(2.0));
  if ( dx != (Int)pow(2.0,m_shift_div) ) {
    printf("TComLut3D:\tnbp should be power of 2.\n");
    exit(0);
  }
  m_shift_div3  = 3 * m_shift_div;

  m_shift_divY  = m_shift_div + (internalBitDepthY - 8);
  m_shift_divC  = m_shift_div + (internalBitDepthC - 8);

  m_shift_div3Y = m_shift_divY + 2*m_shift_divC - (internalBitDepthY - 8);
  m_shift_div3C = m_shift_divY + 2*m_shift_divC - (internalBitDepthC - 8);
#else
  printf("TComLut3D::set_ibd() not implemented.\n");
  exit(0);
#endif
}
#endif

Void  TComLut3D::reset( Bool bResetNode, Bool bResetFlag, Bool bUsedFlag )
{
  int iLayer  = getNbLayer(m_nbp, 8) - 1;

  while ( iLayer >= 0 ) {
    int di = (m_nbp-1) / (1<<iLayer);
    int dj = (m_nbp-1) / (1<<iLayer);
    int dk = (m_nbp-1) / (1<<iLayer);
    for ( int i=0 ; i<m_nbp ; i+=di )
    for ( int j=0 ; j<m_nbp ; j+=dj )
      for ( int k=0 ; k<m_nbp ; k+=dk ) {
        node* n = getNode(i,j,k);
        n->iLayer    = iLayer;
        if ( bResetFlag ) {
          n->bEncoded = false;
          n->bUsed    = bUsedFlag;
        }
        if ( bResetNode ) {
          n->sY        = 0;
          n->sU        = 0;
          n->sV        = 0;
        }
      }

    iLayer--;
  }

}

Void  TComLut3D::setDefault()
{
  int dr = BOX_MAX / (m_nbp-1);
  int dg = BOX_MAX / (m_nbp-1);
  int db = BOX_MAX / (m_nbp-1);

  for ( Int i=0 ; i<m_nbp ; i++ )
  for ( Int j=0 ; j<m_nbp ; j++ )
  for ( Int k=0 ; k<m_nbp ; k++ ) {
    node  *n = getNode(i,j,k);
    n->sY = i * dr;
    n->sU = j * dg;
    n->sV = k * db;
  }
}

// Tri-Linear interpolation :
Void  TComLut3D::interpol( node *nodeIn, node *nodeOut )
{
#if USE_SHIFT_DIV
  int i = nodeIn->sY >> m_shift_div;
  int j = nodeIn->sU >> m_shift_div;
  int k = nodeIn->sV >> m_shift_div;

  int er  = nodeIn->sY - (i << m_shift_div);
  int eg  = nodeIn->sU - (j << m_shift_div);
  int eb  = nodeIn->sV - (k << m_shift_div);

  int si[2], sj[2], sk[2];
  si[0] = (m_dr - er);
  si[1] = (er);
  sj[0] = (m_dg - eg);
  sj[1] = (eg);
  sk[0] = (m_db - eb);
  sk[1] = (eb);
#else
  int dr = BOX_MAX / (m_nbp-1);
  int dg = BOX_MAX / (m_nbp-1);
  int db = BOX_MAX / (m_nbp-1);

  int i = nodeIn->sY / dr;
  int j = nodeIn->sU / dg;
  int k = nodeIn->sV / db;

  int er  = nodeIn->sY - i*dr;
  int eg  = nodeIn->sU - j*dg;
  int eb  = nodeIn->sV - k*db;

  int si[2], sj[2], sk[2];
  si[0] = (dr - er);
  si[1] = (er);
  sj[0] = (dg - eg);
  sj[1] = (eg);
  sk[0] = (db - eb);
  sk[1] = (eb);
#endif

  UInt64 out_Y = 0;
  UInt64 out_U = 0;
  UInt64 out_V = 0;
  for ( int l=0 ; l<2 ; l++ )
  for ( int m=0 ; m<2 ; m++ )
    for ( int n=0 ; n<2 ; n++ ) {
      Int   iWeight = si[l]*sj[m]*sk[n];
      //if ( !iWeight ) continue; // does it really accelerate ?

      node  *p      = getNode( i+l, j+m, k+n );

      out_Y += iWeight * p->sY;
      out_U += iWeight * p->sU;
      out_V += iWeight * p->sV;

      p->bUsed  = true;
    }

#if USE_SHIFT_DIV
  UInt64 offset = 1 << (m_shift_div3-1);
  nodeOut->sY   = (Short)( (out_Y + offset) >> m_shift_div3 );
  nodeOut->sU   = (Short)( (out_U + offset) >> m_shift_div3 );
  nodeOut->sV   = (Short)( (out_V + offset) >> m_shift_div3 );
#else
  Int   denom   = dr*dg*db;
  UInt64 offset  = denom >> 1;
  nodeOut->sY   = (Short)( (out_Y + offset) / denom );
  nodeOut->sU   = (Short)( (out_U + offset) / denom );
  nodeOut->sV   = (Short)( (out_V + offset) / denom );
#endif
}

#if USE_IBD
Void  TComLut3D::interpol_ibd( node *nodeIn, node *nodeOut )
{
#if USE_SHIFT_DIV
  int i = nodeIn->sY >> m_shift_divY;
  int j = nodeIn->sU >> m_shift_divC;
  int k = nodeIn->sV >> m_shift_divC;

  int er  = nodeIn->sY - (i << m_shift_divY);
  int eg  = nodeIn->sU - (j << m_shift_divC);
  int eb  = nodeIn->sV - (k << m_shift_divC);

  int si[2], sj[2], sk[2];
  si[0] = (m_dr - er);
  si[1] = (er);
  sj[0] = (m_dg - eg);
  sj[1] = (eg);
  sk[0] = (m_db - eb);
  sk[1] = (eb);

  UInt64 out_Y = 0;
  UInt64 out_U = 0;
  UInt64 out_V = 0;
  for ( int l=0 ; l<2 ; l++ )
  for ( int m=0 ; m<2 ; m++ )
    for ( int n=0 ; n<2 ; n++ ) {
      Int   iWeight = si[l]*sj[m]*sk[n];
      //if ( !iWeight ) continue; // does it really accelerate ?

      node  *p      = getNode( i+l, j+m, k+n );

      out_Y += iWeight * p->sY;
      out_U += iWeight * p->sU;
      out_V += iWeight * p->sV;

      p->bUsed  = true;
    }

  UInt64 offsetY  = 1 << (m_shift_div3Y-1);
  UInt64 offsetC  = 1 << (m_shift_div3C-1);
  nodeOut->sY     = (Short)( (out_Y + offsetY) >> m_shift_div3Y );
  nodeOut->sU     = (Short)( (out_U + offsetC) >> m_shift_div3C );
  nodeOut->sV     = (Short)( (out_V + offsetC) >> m_shift_div3C );
#else
  printf("TComLut3D::interpol_ibd() not implemented.\n");
  exit(0);
#endif
}
#endif

// Tetrahedral interpolation :
Void  TComLut3D::interpol_tetra( node *nodeIn, node *nodeOut )
{
#if USE_SHIFT_DIV
  int i = nodeIn->sY >> m_shift_div;
  int j = nodeIn->sU >> m_shift_div;
  int k = nodeIn->sV >> m_shift_div;

  int er  = nodeIn->sY - (i << m_shift_div);
  int eg  = nodeIn->sU - (j << m_shift_div);
  int eb  = nodeIn->sV - (k << m_shift_div);
#else
  int dr = BOX_MAX / (m_nbp-1);
  int dg = BOX_MAX / (m_nbp-1);
  int db = BOX_MAX / (m_nbp-1);

  int i = nodeIn->sY / dr;
  int j = nodeIn->sU / dg;
  int k = nodeIn->sV / db;

  int er  = nodeIn->sY - i*dr;
  int eg  = nodeIn->sU - j*dg;
  int eb  = nodeIn->sV - k*db;
#endif

 
  UInt64 out_Y = 0;
  UInt64 out_U = 0;
  UInt64 out_V = 0;
  
  node  nodeArray[8];
  
  int inode = 0;
  for ( int l=0 ; l<2 ; l++ )
  for ( int m=0 ; m<2 ; m++ )
    for ( int n=0 ; n<2 ; n++ ) {
      

      node  *p      = getNode( i+l, j+m, k+n );
     
      nodeArray[inode].sY = p ->sY;
      nodeArray[inode].sU = p ->sU;
      nodeArray[inode].sV = p ->sV;
      
      p->bUsed  = true;
      inode++;
    }


#if USE_SHIFT_DIV
    if ((er >= eg) && (eg >= eb)){
      //node  *nodeArray_0 = getNode( i,   j, k );
      //node  *nodeArray_4 = getNode( i+1, j, k );

      out_Y = ((nodeArray[0].sY) << m_shift_div) + er * (nodeArray[4].sY - nodeArray[0].sY) + eg *(nodeArray[6].sY - nodeArray[4].sY)+ eb *(nodeArray[7].sY - nodeArray[6].sY);
      out_U = ((nodeArray[0].sU) << m_shift_div) + er * (nodeArray[4].sU - nodeArray[0].sU) + eg *(nodeArray[6].sU - nodeArray[4].sU)+ eb *(nodeArray[7].sU - nodeArray[6].sU);
      out_V = ((nodeArray[0].sV) << m_shift_div) + er * (nodeArray[4].sV - nodeArray[0].sV) + eg *(nodeArray[6].sV - nodeArray[4].sV)+ eb *(nodeArray[7].sV - nodeArray[6].sV);
    }
    else if((er > eb) && (eb >= eg)){
      out_Y = ((nodeArray[0].sY) << m_shift_div) + er * (nodeArray[4].sY - nodeArray[0].sY) + eg *(nodeArray[7].sY - nodeArray[5].sY)+ eb *(nodeArray[5].sY - nodeArray[4].sY);
      out_U = ((nodeArray[0].sU) << m_shift_div) + er * (nodeArray[4].sU - nodeArray[0].sU) + eg *(nodeArray[7].sU - nodeArray[5].sU)+ eb *(nodeArray[5].sU - nodeArray[4].sU);
      out_V = ((nodeArray[0].sV) << m_shift_div) + er * (nodeArray[4].sV - nodeArray[0].sV) + eg *(nodeArray[7].sV - nodeArray[5].sV)+ eb *(nodeArray[5].sV - nodeArray[4].sV);
    }
    else if((eb >= er) && (er > eg)){
      out_Y = ((nodeArray[0].sY) << m_shift_div) + er * (nodeArray[5].sY - nodeArray[1].sY) + eg *(nodeArray[7].sY - nodeArray[5].sY)+ eb *(nodeArray[1].sY - nodeArray[0].sY);
      out_U = ((nodeArray[0].sU) << m_shift_div) + er * (nodeArray[5].sU - nodeArray[1].sU) + eg *(nodeArray[7].sU - nodeArray[5].sU)+ eb *(nodeArray[1].sU - nodeArray[0].sU);
      out_V = ((nodeArray[0].sV) << m_shift_div) + er * (nodeArray[5].sV - nodeArray[1].sV) + eg *(nodeArray[7].sV - nodeArray[5].sV)+ eb *(nodeArray[1].sV - nodeArray[0].sV);
    }
    else if((eg > er) && (er >= eb)){
      out_Y = ((nodeArray[0].sY) << m_shift_div) + er * (nodeArray[6].sY - nodeArray[2].sY) + eg *(nodeArray[2].sY - nodeArray[0].sY)+ eb *(nodeArray[7].sY - nodeArray[6].sY);
      out_U = ((nodeArray[0].sU) << m_shift_div) + er * (nodeArray[6].sU - nodeArray[2].sU) + eg *(nodeArray[2].sU - nodeArray[0].sU)+ eb *(nodeArray[7].sU - nodeArray[6].sU);
      out_V = ((nodeArray[0].sV) << m_shift_div) + er * (nodeArray[6].sV - nodeArray[2].sV) + eg *(nodeArray[2].sV - nodeArray[0].sV)+ eb *(nodeArray[7].sV - nodeArray[6].sV);
    }
    else if((eg > eb) && (eb > er)){
      out_Y = ((nodeArray[0].sY) << m_shift_div) + er * (nodeArray[7].sY - nodeArray[3].sY) + eg *(nodeArray[2].sY - nodeArray[0].sY)+ eb *(nodeArray[3].sY - nodeArray[2].sY);
      out_U = ((nodeArray[0].sU) << m_shift_div) + er * (nodeArray[7].sU - nodeArray[3].sU) + eg *(nodeArray[2].sU - nodeArray[0].sU)+ eb *(nodeArray[3].sU - nodeArray[2].sU);
      out_V = ((nodeArray[0].sV) << m_shift_div) + er * (nodeArray[7].sV - nodeArray[3].sV) + eg *(nodeArray[2].sV - nodeArray[0].sV)+ eb *(nodeArray[3].sV - nodeArray[2].sV);
    }
    else if((eb >= eg) && (eg >= er)){
      out_Y = ((nodeArray[0].sY) << m_shift_div) + er * (nodeArray[7].sY - nodeArray[3].sY) + eg *(nodeArray[3].sY - nodeArray[1].sY)+ eb *(nodeArray[1].sY - nodeArray[0].sY);
      out_U = ((nodeArray[0].sU) << m_shift_div) + er * (nodeArray[7].sU - nodeArray[3].sU) + eg *(nodeArray[3].sU - nodeArray[1].sU)+ eb *(nodeArray[1].sU - nodeArray[0].sU);
      out_V = ((nodeArray[0].sV) << m_shift_div) + er * (nodeArray[7].sV - nodeArray[3].sV) + eg *(nodeArray[3].sV - nodeArray[1].sV)+ eb *(nodeArray[1].sV - nodeArray[0].sV);
    }
    else
       printf("\n\n **** ERROR (TetraHedral) ****\n");
#else
    if ((er >= eg) && (eg >= eb)){
      out_Y = (nodeArray[0].sY)* dr + er * (nodeArray[4].sY - nodeArray[0].sY) + eg *(nodeArray[6].sY - nodeArray[4].sY)+ eb *(nodeArray[7].sY - nodeArray[6].sY);
      out_U = (nodeArray[0].sU)* dg + er * (nodeArray[4].sU - nodeArray[0].sU) + eg *(nodeArray[6].sU - nodeArray[4].sU)+ eb *(nodeArray[7].sU - nodeArray[6].sU);
      out_V = (nodeArray[0].sV)* db + er * (nodeArray[4].sV - nodeArray[0].sV) + eg *(nodeArray[6].sV - nodeArray[4].sV)+ eb *(nodeArray[7].sV - nodeArray[6].sV);
    }
    else if((er > eb) && (eb >= eg)){
      out_Y = (nodeArray[0].sY)* dr + er * (nodeArray[4].sY - nodeArray[0].sY) + eg *(nodeArray[7].sY - nodeArray[5].sY)+ eb *(nodeArray[5].sY - nodeArray[4].sY);
      out_U = (nodeArray[0].sU)* dg + er * (nodeArray[4].sU - nodeArray[0].sU) + eg *(nodeArray[7].sU - nodeArray[5].sU)+ eb *(nodeArray[5].sU - nodeArray[4].sU);
      out_V = (nodeArray[0].sV)* db + er * (nodeArray[4].sV - nodeArray[0].sV) + eg *(nodeArray[7].sV - nodeArray[5].sV)+ eb *(nodeArray[5].sV - nodeArray[4].sV);
    }
    else if((eb >= er) && (er > eg)){
      out_Y = (nodeArray[0].sY)* dr + er * (nodeArray[5].sY - nodeArray[1].sY) + eg *(nodeArray[7].sY - nodeArray[5].sY)+ eb *(nodeArray[1].sY - nodeArray[0].sY);
      out_U = (nodeArray[0].sU)* dg + er * (nodeArray[5].sU - nodeArray[1].sU) + eg *(nodeArray[7].sU - nodeArray[5].sU)+ eb *(nodeArray[1].sU - nodeArray[0].sU);
      out_V = (nodeArray[0].sV)* db + er * (nodeArray[5].sV - nodeArray[1].sV) + eg *(nodeArray[7].sV - nodeArray[5].sV)+ eb *(nodeArray[1].sV - nodeArray[0].sV);
    }
    else if((eg > er) && (er >= eb)){
      out_Y = (nodeArray[0].sY)* dr + er * (nodeArray[6].sY - nodeArray[2].sY) + eg *(nodeArray[2].sY - nodeArray[0].sY)+ eb *(nodeArray[7].sY - nodeArray[6].sY);
      out_U = (nodeArray[0].sU)* dg + er * (nodeArray[6].sU - nodeArray[2].sU) + eg *(nodeArray[2].sU - nodeArray[0].sU)+ eb *(nodeArray[7].sU - nodeArray[6].sU);
      out_V = (nodeArray[0].sV)* db + er * (nodeArray[6].sV - nodeArray[2].sV) + eg *(nodeArray[2].sV - nodeArray[0].sV)+ eb *(nodeArray[7].sV - nodeArray[6].sV);
    }
    else if((eg > eb) && (eb > er)){
      out_Y = (nodeArray[0].sY)* dr + er * (nodeArray[7].sY - nodeArray[3].sY) + eg *(nodeArray[2].sY - nodeArray[0].sY)+ eb *(nodeArray[3].sY - nodeArray[2].sY);
      out_U = (nodeArray[0].sU)* dg + er * (nodeArray[7].sU - nodeArray[3].sU) + eg *(nodeArray[2].sU - nodeArray[0].sU)+ eb *(nodeArray[3].sU - nodeArray[2].sU);
      out_V = (nodeArray[0].sV)* db + er * (nodeArray[7].sV - nodeArray[3].sV) + eg *(nodeArray[2].sV - nodeArray[0].sV)+ eb *(nodeArray[3].sV - nodeArray[2].sV);
    }
    else if((eb >= eg) && (eg >= er)){
      out_Y = (nodeArray[0].sY)* dr + er * (nodeArray[7].sY - nodeArray[3].sY) + eg *(nodeArray[3].sY - nodeArray[1].sY)+ eb *(nodeArray[1].sY - nodeArray[0].sY);
      out_U = (nodeArray[0].sU)* dg + er * (nodeArray[7].sU - nodeArray[3].sU) + eg *(nodeArray[3].sU - nodeArray[1].sU)+ eb *(nodeArray[1].sU - nodeArray[0].sU);
      out_V = (nodeArray[0].sV)* db + er * (nodeArray[7].sV - nodeArray[3].sV) + eg *(nodeArray[3].sV - nodeArray[1].sV)+ eb *(nodeArray[1].sV - nodeArray[0].sV);
    }
    else
       printf("\n\n **** ERROR (TetraHedral) ****\n");
#endif
  
  UInt64 offset  = 0;
#if USE_SHIFT_DIV
  nodeOut->sY   = (int)( (out_Y + offset) >> m_shift_div );
  nodeOut->sU   = (int)( (out_U + offset) >> m_shift_div );
  nodeOut->sV   = (int)( (out_V + offset) >> m_shift_div );
#else
  nodeOut->sY   = (int)( (out_Y + offset) / dr );
  nodeOut->sU   = (int)( (out_U + offset) / dg );
  nodeOut->sV   = (int)( (out_V + offset) / db );
#endif 
}


Void  TComLut3D::interpolLut_fromUpperLayer( node *nodeIn, node *nodeOut, Int iUpperLayer )
{
  int dr = BOX_MAX / (m_nbp-1);
  int dg = BOX_MAX / (m_nbp-1);
  int db = BOX_MAX / (m_nbp-1);
  int Dr = BOX_MAX / (1 << iUpperLayer);
  int Dg = BOX_MAX / (1 << iUpperLayer);
  int Db = BOX_MAX / (1 << iUpperLayer);

  int nbp1  = (m_nbp-1) / (1 << iUpperLayer);
  int nbp2  = nbp1 >> 1;
  int i = (nodeIn->sY / Dr) * nbp1;
  int j = (nodeIn->sU / Dg) * nbp1;
  int k = (nodeIn->sV / Db) * nbp1;

  int er  = (nodeIn->sY - i*dr) / nbp2;
  int eg  = (nodeIn->sU - j*dg) / nbp2;
  int eb  = (nodeIn->sV - k*db) / nbp2;
  Dr /= nbp2;
  Dg /= nbp2;
  Db /= nbp2;

  int si[2], sj[2], sk[2];
  si[0] = (Dr - er);
  si[1] = (er);
  sj[0] = (Dg - eg);
  sj[1] = (eg);
  sk[0] = (Db - eb);
  sk[1] = (eb);

  UInt64 out_Y = 0;
  UInt64 out_U = 0;
  UInt64 out_V = 0;
  for ( int l=0 ; l<2 ; l++ )
  for ( int m=0 ; m<2 ; m++ )
    for ( int n=0 ; n<2 ; n++ ) {

      int wi  = si[l]*sj[m]*sk[n];
      if ( !wi ) continue;

      node  *P = getNode( i+l*nbp1, j+m*nbp1, k+n*nbp1 );

      out_Y += wi * P->sY;
      out_U += wi * P->sU;
      out_V += wi * P->sV;
    }

#if USE_SHIFT_DIV
  Int shift_div = 18;
  UInt64 offset = 1 << (shift_div-1);
  nodeOut->sY = (int)( (out_Y + offset) >> shift_div );
  nodeOut->sU = (int)( (out_U + offset) >> shift_div );
  nodeOut->sV = (int)( (out_V + offset) >> shift_div );
#else
  Int   denom   = Dr*Dg*Db;
  UInt64 offset  = denom >> 1;
  nodeOut->sY = (int)( (out_Y + offset) / denom );
  nodeOut->sU = (int)( (out_U + offset) / denom );
  nodeOut->sV = (int)( (out_V + offset) / denom );
#endif
}

Void  TComLut3D::encode(Int errMax, Int errQ, TComLut3D *lutRef, TComLut3D *lutResiduals, Bool bVerbose )
{
  encode(0, 0,0,0, errMax, errQ, lutRef, lutResiduals, m_octTree, NULL, bVerbose );
}

Void  TComLut3D::encode(Int iLayer, int r0, int g0, int b0, int errMax, Int errQ, TComLut3D *lutRef, TComLut3D *lutResiduals, octTree *pOctTree, octTree *pOctTreeParent, Bool bVerbose )
{
  int dr = BOX_MAX / (m_nbp-1);
  int dg = BOX_MAX / (m_nbp-1);
  int db = BOX_MAX / (m_nbp-1);
  int Dr = BOX_MAX / (1 << iLayer);
  int Dg = BOX_MAX / (1 << iLayer);
  int Db = BOX_MAX / (1 << iLayer);

  node  *nodeRef;
  node  nodePred;
  node  nodeIn;

  //printf("\nencodeCube(layer %d)\n", iLayer);

  // encode current cube (iLayer)
  for ( int r=r0+Dr ; r>=r0 ; r-=Dr ) {
    for ( int g=g0+Dg ; g>=g0 ; g-=Dg ) {
      for ( int b=b0+Db ; b>=b0 ; b-=Db ) {

        int i = r / dr;
        int j = g / dg;
        int k = b / db;

        nodeIn.sY = r;
        nodeIn.sU = g;
        nodeIn.sV = b;

        nodeRef = lutRef->getNode(i,j,k);

        if ( !getNode(i,j,k)->bEncoded ) {  // not yet encoded
          nodePred.iLayer = iLayer;

          // get predictor :
          if ( iLayer ) interpolLut_fromUpperLayer( &nodeIn, &nodePred, iLayer-1 );
          else {
            nodePred.sY = r;
            nodePred.sU = g;
            nodePred.sV = b;
          }

          node  *nodeCoded    = getNode(i,j,k);

          node*  nodeResidue  = lutResiduals->getNode(i,j,k);

          Int err=0;
          if ( nodeRef->bUsed )
            err = getNodeError( nodeRef, &nodePred );

          // encode the difference :
          if ( err > errMax ) {
            Int offset  = errQ >> 1;

            nodeResidue->sY  = ( nodeRef->sY - nodePred.sY + offset ) / errQ;
            nodeResidue->sU  = ( nodeRef->sU - nodePred.sU + offset ) / errQ;
            nodeResidue->sV  = ( nodeRef->sV - nodePred.sV + offset ) / errQ;

            if ( nodeResidue->sY || nodeResidue->sU || nodeResidue->sV ) {
              nodeResidue->bEncoded = true;
            }

            nodeCoded->sY  = nodePred.sY + nodeResidue->sY*errQ;
            nodeCoded->sU  = nodePred.sU + nodeResidue->sU*errQ;
            nodeCoded->sV  = nodePred.sV + nodeResidue->sV*errQ;

            if ( bVerbose )
              printf("L%d-node[%d][%d][%d]:\t(%d;%d;%d) = P(%d;%d;%d) + R(%d;%d;%d)\n", iLayer, r,g,b, 
                nodeCoded->sY,nodeCoded->sU,nodeCoded->sV,
                nodePred.sY,nodePred.sU,nodePred.sV,
                nodeResidue->sY,nodeResidue->sU,nodeResidue->sV);

            if ( pOctTreeParent && nodeCoded->iLayer == iLayer )  pOctTreeParent->bSplit = true;
          }
          else {
            copyNode( nodeCoded, &nodePred );

            if ( bVerbose )
              printf("L%d-node[%d][%d][%d]:\t(%d;%d;%d) = P(%d;%d;%d)\n", iLayer, r,g,b, 
                nodeCoded->sY,nodeCoded->sU,nodeCoded->sV,
                nodePred.sY,nodePred.sU,nodePred.sV);

            nodeResidue->bEncoded = false;
          }

          nodeCoded->bEncoded   = true;

        }

      }
    }
  }

  // Recursively encode cubes/octants:
  if ( bVerbose ) printf("\n");

  //if ( !bWriteStream || (bWriteStream && pOctTree->bSplit) )
  if ( iLayer < getNbLayer(m_nbp, 8)-1 ) {
    Dr /= 2;
    Dg /= 2;
    Db /= 2;

    // encode cubes in lower layer (iLayer+1)
    int iLeave=0;
    for ( int r=r0+Dr ; r>=r0 ; r-=Dr ) {
      for ( int g=g0+Dg ; g>=g0 ; g-=Dg ) {
        for ( int b=b0+Db ; b>=b0 ; b-=Db , iLeave++ ) {
          encode(iLayer+1, r, g, b, errMax, errQ, lutRef, lutResiduals, (octTree *)pOctTree->leave[iLeave], pOctTree, bVerbose );

          if ( pOctTreeParent && pOctTree->bSplit ) pOctTreeParent->bSplit = true;
        }
      }
    }
  }

}

Void  TComLut3D::decode( Int iLayer, int r0, int g0, int b0, Int errQ, TComLut3D* lutResiduals, octTree *pOctTree, Bool bVerbose )
{
  //printf("TComLut3D::decode()...\n");

  int dr = BOX_MAX / (m_nbp-1);
  int dg = BOX_MAX / (m_nbp-1);
  int db = BOX_MAX / (m_nbp-1);
  int Dr = BOX_MAX / (1 << iLayer);
  int Dg = BOX_MAX / (1 << iLayer);
  int Db = BOX_MAX / (1 << iLayer);

  node  nodePred;
  node  nodeIn;

  //printf("\ndecodeCube(layer %d)\n", iLayer);

  // encode current cube (iLayer)
  for ( int r=r0+Dr ; r>=r0 ; r-=Dr ) {
    for ( int g=g0+Dg ; g>=g0 ; g-=Dg ) {
      for ( int b=b0+Db ; b>=b0 ; b-=Db ) {

        int i = r / dr;
        int j = g / dg;
        int k = b / db;

        nodeIn.sY = r;
        nodeIn.sU = g;
        nodeIn.sV = b;

        node  *nodeResiduals  = lutResiduals->getNode(i,j,k);

        nodePred.iLayer = iLayer;

        // get predictor :
        if ( iLayer ) interpolLut_fromUpperLayer( &nodeIn, &nodePred, iLayer-1 );
        else {
          nodePred.sY = r;
          nodePred.sU = g;
          nodePred.sV = b;
        }

        node  *nodeCoded  = getNode(i,j,k);

        if ( !nodeCoded->bEncoded ) {  // not yet encoded

          nodeCoded->sY = nodePred.sY + nodeResiduals->sY * errQ;
          nodeCoded->sU = nodePred.sU + nodeResiduals->sU * errQ;
          nodeCoded->sV = nodePred.sV + nodeResiduals->sV * errQ;

          if ( bVerbose )
            printf("L%d-node[%d][%d][%d]:\t(%d;%d;%d) = P(%d;%d;%d) + R(%d;%d;%d)\n", iLayer, r,g,b, 
              nodeCoded->sY,nodeCoded->sU,nodeCoded->sV,
              nodePred.sY,nodePred.sU,nodePred.sV,
              nodeResiduals->sY,nodeResiduals->sU,nodeResiduals->sV);

          nodeCoded->bEncoded   = true;
        }

      }
    }
  }

  // Recursively encode cubes/octants:
  if ( bVerbose ) printf("\n");

  //if ( pOctTree->bSplit )
  if ( iLayer < getNbLayer(m_nbp, 8)-1 ) {
    Dr /= 2;
    Dg /= 2;
    Db /= 2;

    // encode cubes in lower layer (iLayer+1)
    int iLeave=0;
    for ( int r=r0+Dr ; r>=r0 ; r-=Dr ) {
      for ( int g=g0+Dg ; g>=g0 ; g-=Dg ) {
        for ( int b=b0+Db ; b>=b0 ; b-=Db , iLeave++ ) {
          decode(iLayer+1, r, g, b, errQ, lutResiduals, (octTree *)pOctTree->leave[iLeave], bVerbose );
        }
      }
    }
  }

}

Void  TComLut3D::prune( Bool bVerbose )
{
  if ( bVerbose ) printf("TComLut3D::prune()...\n");

  int dr = BOX_MAX / (m_nbp-1);
  int dg = BOX_MAX / (m_nbp-1);
  int db = BOX_MAX / (m_nbp-1);

  for ( int r=0 ; r<=BOX_MAX ; r+=dr )
  for ( int g=0 ; g<=BOX_MAX ; g+=dg )
  for ( int b=0 ; b<=BOX_MAX ; b+=db ) {
    int i = r / dr;
    int j = g / dg;
    int k = b / db;
    node  *n  = getNode(i,j,k);
    if ( bVerbose )
      if ( !n->bUsed )
        printf("TComLut3D::prune():\tL%d-node[%d][%d][%d] not used\n",  n->iLayer, r,g,b );
      else
        printf("TComLut3D::prune():\tL%d-node[%d][%d][%d] used\n",      n->iLayer, r,g,b );
  }
}

